% Perform face relighting based on reference images
%
% Jiansheng Chen, Guangda Su, Jinping He, Shenglan Ben, "Face Image 
% Relighting using Locally Constrained Global Optimization", ECCV'10
%
%01-10-10, Jiansheng Chen

% result = img*imgRefA/imgRefB
function relight(img, pt, imgRefA, ptRefA, imgRefB, ptRefB)

    % morphing the reference images
    scaleA = max(size(img,1)/size(imgRefA,1),size(img,2)/size(imgRefA,2));
    imgRefA = imresize(imgRefA, scaleA);    ptRefA = ptRefA*scaleA;
    scaleB =  max(size(img,1)/size(imgRefB,1),size(img,2)/size(imgRefB,2));
    imgRefB = imresize(imgRefB,scaleB);     ptRefB = ptRefB*scaleB;
    try
        imgRefA = rgb2gray(imgRefA);
        imgRefB = rgb2gray(imgRefB);
    catch
    end
    imgRefA = morph(imgRefA, ptRefA, pt); imgRefA = imgRefA(1:size(img,1), 1:size(img,2));
    imgRefB = morph(imgRefB, ptRefB, pt); imgRefB = imgRefB(1:size(img,1), 1:size(img,2));
    
    % create the mask of convex hull of pt
    [X,Y] = meshgrid(1:size(img, 2), 1:size(img, 1));
    ch = convhulln(pt);
    mask = inpolygon(X,Y,pt(ch(:,1),1),pt(ch(:,1),2));
    margin = 5;
    xpro = sum(mask, 1); ypro = sum(mask, 2);
    left = max(min(find(xpro>0))-margin, 0); right = min(max(find(xpro>0))+margin, size(mask, 2));
    up = max(min(find(ypro>0))-margin, 0); down = min(max(find(ypro>0))+margin, size(mask, 1));
    rect = [left, up, right-left+1, down-up+1];
    mask = imcrop(mask, rect);
    imgRefA = imcrop(imgRefA, rect).*mask; 
    imgRefB = imcrop(imgRefB, rect).*mask; 
    hsv_image = rgb2hsv(imcrop(img, rect));
    hsv_image(:,:,1) = hsv_image(:,:,1).*mask;
    hsv_image(:,:,2) = hsv_image(:,:,2).*mask;
    hsv_image(:,:,3) = hsv_image(:,:,3).*mask;
    img_crop = hsv2rgb(hsv_image);
    
    % precalculate some of the temp varibles
    win = 3; % size of the shifting window, make it an ODD number!
    [H, W] = size(hsv_image(:,:,3));
    sigma2_matrix = zeros(H-win+1, W-win+1);
    for i = 1:H-win+1
        for j = 1:W-win+1
            sub1 = hsv_image(i:i+win-1, j:j+win-1, 3);
            sigma2_matrix(i,j) = sum(sum(sub1.^2));
        end
    end
    
    % calculate the ratio image and remove some of the spikes
    C = (filter2(fspecial('average', win), imgRefA, 'same')+0.001)./(filter2(fspecial('average', win), imgRefB, 'same')+0.001);
    C = imopen(C, strel('disk', 2)); C(C<0) = 0;
    
    % perform illumination transform using locally constrained global optimization
    hsv_image(:,:,3) = logo(hsv_image(:,:,3), C, win, sigma2_matrix);
    figure('name', 'Face Image Relighting');
    subplot(1,2,1); imshow(img_crop); title('Orignal Image');
    subplot(1,2,2);  imshow(hsv2rgb(hsv_image)); title('Relighted Image');
    
end